// RCM - Written by Nauful

// License for this source file:
// Commercial use: Not allowed without prior permission
// Distribution, modifcation, private use: Allowed if the original source is credited and modifications are not represented as part of the original source
// No guarantee of correctness is provided beyond a best-effort basis, no liability is assumed

#define _CRT_DISABLE_PERFCRIT_LOCKS

#include <intrin.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#pragma warning(disable: 4267) // conversion from 'x' to 'y', possible loss of data
#pragma warning(disable: 4804) // '<<=' : unsafe use of type 'bool' in operation

#ifdef DEBUG
#	ifdef _MSC_VER
#		define _CRTDBG_MAP_ALLOC
#		include <crtdbg.h>
#		ifndef ASSERT
#			define ASSERT(x) \
			{ \
			if (!!!(x)) \
				{ \
				printf("Assert failed!\nCondition: %s\n%s:%d\n", #x, __FILE__, __LINE__); \
				__debugbreak(); \
				} \
			}
#		endif
#	else
#			ifndef ASSERT
#			define ASSERT(x) \
			{ \
			if (!!!(x)) \
				{ \
				printf("Assert failed\nCondition: %s\n%s:%d\n", #x, __FILE__, __LINE__); \
				} \
			}
#		endif
#	endif
#	ifndef ASSERT_DEBUG
#		define ASSERT_DEBUG(x) ASSERT(x)
#	endif
#else
#	ifndef ASSERT
#		define ASSERT(x) \
		{ \
		if (!!!(x)) \
			{ \
			printf("Assert failed\n%s:%d\n", __FILE__, __LINE__); \
			} \
		}
#	endif
#	ifndef ASSERT_DEBUG
#		define ASSERT_DEBUG(x) ASSERT(x)
#	endif
#endif

#ifdef _MSC_VER
#	ifndef ALIGN
#		define ALIGN(x) __declspec(align(x))
#	endif
#	ifndef RESTRICT
#		define RESTRICT __restrict
#	endif
#	ifndef INLINE
#		define INLINE __forceinline
#	endif
#else
#	ifndef ALIGN
#		define ALIGN(x) __attribute__((aligned(x)))
#	endif
#	ifndef RESTRICT
#		define RESTRICT __restrict
#	endif
#	ifndef INLINE
#		define INLINE inline
#	endif
#endif

#ifndef countof
#	define countof(x) (sizeof((x)) / sizeof((x)[0]))
#endif

#ifndef MIN
#	define MIN(x, y) (((x) <= (y) ? (x) : (y)))
#endif

#ifndef MAX
#	define MAX(x, y) (((x) >= (y) ? (x) : (y)))
#endif

#ifdef x64
#	ifdef _MSC_VER
#		define fseek _fseeki64
#		define ftell _ftelli64
#	else
#		define fseek fseeko64
#		define ftell ftello64
#	endif
#endif

typedef unsigned int uint;
typedef unsigned char byte;

typedef unsigned char uint8;
typedef unsigned short uint16;
typedef unsigned int uint32;
typedef unsigned long long uint64;

typedef char int8;
typedef short int16;
typedef int int32;
typedef long long int64;

#ifdef x64
typedef uint64 uintptr;
#else
typedef uint32 uintptr;
#endif

const static int WINDOW_BITS = 16;
const static int WINDOW_SIZE = 1 << WINDOW_BITS;
const static int WINDOW_MASK = WINDOW_SIZE - 1;
const static int WINDOW_DOUBLE_MASK = 2 * WINDOW_SIZE - 1;

const static int WINDOW_SIZE_WORST = 64 + 1.01 * WINDOW_SIZE;
const static int FAST_COPY_PADDING = 32;

const static int RCM_MATCH_MIN = 4;
const static int RCM_MATCH_MED_MAX = (1 << 6) + RCM_MATCH_MIN;
const static int RCM_MATCH_LONG_MAX = (1 << (6 + 8)) + RCM_MATCH_MIN;
const static int RCM_MATCH_SEMIOPTIMAL_MAX = 256;

const static int RCM_MATCH_SEARCH_PADDING = 32;

const static int RCM_INCOMPRESSIBLE_STEP = 16;
const static int RCM_INCOMPRESSIBLE_MAX = 64;

const static int RCM_OFFSET_THRESHOLD = 1 << 11;
const static int RCM_MATCH_THRESHOLD = RCM_MATCH_MIN + 8;

template<int HASH_BITS>
int hash4(const byte* p)
{
	const int HASH_MASK = (1 << HASH_BITS) - 1;

	uint32 h = (*reinterpret_cast<const uint32*>(p) * 2654435761u) >> (32 - HASH_BITS);
	return h & HASH_MASK;
}

template<bool IS_SAME_BUFFER>
static INLINE void string_copy(const byte* RESTRICT src, byte* RESTRICT dest, int len)
{
#ifdef x64
	if (IS_SAME_BUFFER)
	{
		// Align
		while (dest - src < 8)
		{
			*(volatile uint64*)dest = *(volatile uint64*)src;
			len -= dest - src;
			dest += dest - src;
		}
	}

	while (len > 0)
	{
		*(volatile uint64*)dest = *(volatile uint64*)src;
		len -= 8;

		dest += 8;
		src += 8;
	}
#else
	if (IS_SAME_BUFFER)
	{
		// Align
		while (dest - src < 4)
		{
			*(volatile uint32*)dest = *(volatile uint32*)src;
			len -= dest - src;
			dest += dest - src;
		}
	}

	while (len > 0)
	{
		*(volatile uint32*)dest = *(volatile uint32*)src;
		len -= 4;

		dest += 4;
		src += 4;
	}
#endif
}

INLINE int match_string(const byte* RESTRICT p1, const byte* RESTRICT p2, int max_len)
{
	int len = 0;
	unsigned long r = 0;

#ifdef x64
	uint64 v = *(uint64*)(p1) ^ *(uint64*)(p2);
	while (!v && len < max_len)
	{
		len += 8;
		v = *(uint64*)(p1 + len) ^ *(uint64*)(p2 + len);
	}

#ifdef _MSC_VER
	if (v && _BitScanForward64(&r, v))
#else
	if (v && (r = __builtin_ctzll(v)))
#endif
	{
		len += r >> 3;
	}
#else
	uint64 v = *(uint32*)(p1) ^ *(uint32*)(p2);
	while (!v && len < max_len)
	{
		len += 4;
		v = *(uint32*)(p1 + len) ^ *(uint32*)(p2 + len);
	}

#ifdef _MSC_VER
	if (v && _BitScanForward(&r, v))
#else
	if (v && (r = __builtin_ctz(v)))
#endif
	{
		len += r >> 3;
	}
#endif

	return len > max_len ? max_len : len;
}

void compress_fast(const char* infilename, const char* outfilename)
{
	const int HASH_BITS = 13;
	const int TRY_BITS = 2;

	const int NUM_TRIES = 1 << TRY_BITS;
	const int HASH_MASK = (1 << HASH_BITS) - 1;
	const int TRY_MASK = (1 << TRY_BITS) - 1;

	FILE* fin = fopen(infilename, "rb");
	if (!fin)
	{
		printf("Could not open %s\n", infilename);
		return;
	}

	FILE* fout = fopen(outfilename, "wb");
	if (!fout)
	{
		fclose(fin);

		printf("Could not open %s\n", outfilename);
		return;
	}

	fseek(fin, 0, SEEK_END);
	uint64 inlen = ftell(fin);
	fseek(fin, 0, SEEK_SET);

	uint16 hash_entry[NUM_TRIES << HASH_BITS];
	byte hash_head[1 << HASH_BITS];
	byte input_mem[2 * WINDOW_SIZE], output[WINDOW_SIZE_WORST];

	memset(hash_entry, 0, sizeof(hash_entry));
	memset(hash_head, 0, sizeof(hash_head));

	memset(input_mem, 0, sizeof(input_mem));
	byte* input = input_mem + WINDOW_SIZE;
	byte* sliding_input = input_mem;

	int n = 0;
	while ((n = fread(input, 1, WINDOW_SIZE, fin)) > 0)
	{
		int rp = 0, wp = 0;
		int incompressible_count = 0, incompressible_step = 0;
		int lit_from = 0;

		while (rp < n)
		{
			int match_lens[2] = { RCM_MATCH_MIN - 1 ,RCM_MATCH_MIN - 1 };
			int match_dists[2] = { 0, 0 };

			int max_match_len = MIN(RCM_MATCH_LONG_MAX - 1, n - rp);
			max_match_len = MAX(0, max_match_len - RCM_MATCH_SEARCH_PADDING);

			if (max_match_len >= RCM_MATCH_MIN && ((rp & incompressible_step) == 0))
			{
				int h = hash4<HASH_BITS>(input + rp);

				for (int i = 0; i < NUM_TRIES; i++)
				{
					int sp = hash_entry[(h << TRY_BITS) + ((hash_head[h] - i) & TRY_MASK)];

					if (sp < rp)
					{
						// From current window
						sp += WINDOW_SIZE;
					}
					else if (sp == rp)
					{
						break;
					}

					if (sliding_input[sp + match_lens[1]] == input[rp + match_lens[1]] &&
						*(uint32*)(sliding_input + sp) == *(uint32*)(input + rp))
					{
						int len = 4 + match_string(sliding_input + sp + 4, input + rp + 4, max_match_len - 4);

						int match_dist = (rp - sp) & WINDOW_MASK;
						if (len > match_lens[0] && match_dist <= RCM_OFFSET_THRESHOLD)
						{
							match_dists[0] = match_dist;
							match_lens[0] = MIN(len, RCM_MATCH_THRESHOLD - 1);
						}

						if (len > match_lens[1])
						{
							match_dists[1] = match_dist;
							match_lens[1] = len;

							if (len == max_match_len)
							{
								break;
							}
						}
					}
				}

				if (match_lens[1] < RCM_MATCH_MIN)
				{
					++incompressible_count;
					if (incompressible_count >= RCM_INCOMPRESSIBLE_STEP &&
						incompressible_step < RCM_INCOMPRESSIBLE_MAX)
					{
						incompressible_step = (incompressible_step << 1) + (incompressible_step == 0);
						incompressible_count = 0;
					}
				}
				else
				{
					incompressible_count = 0;
					incompressible_step >>= 1;
				}
			}

			ASSERT_DEBUG(match_dists[0] <= RCM_OFFSET_THRESHOLD);
			ASSERT_DEBUG(match_lens[0] < RCM_MATCH_THRESHOLD);

			bool match_is_short = false;
			int match_len = 0;
			int match_dist = 0;

			if (match_lens[0] >= RCM_MATCH_MIN)
			{
				match_len = match_lens[0];
				match_dist = match_dists[0];
				match_is_short = true;
			}

			if (match_lens[1] >= RCM_MATCH_MIN && match_lens[1] > match_len + 1)
			{
				match_len = match_lens[1];
				match_dist = match_dists[1];
				match_is_short = false;
			}

			if (match_len >= RCM_MATCH_MIN)
			{
				int sym_lit = rp - lit_from;
				if (sym_lit > 0)
				{
					ASSERT_DEBUG(wp <= sizeof(output));

					byte token = MIN(sym_lit, 63) << 2;
					output[wp++] = token;
					sym_lit -= 63;

					while (sym_lit >= 0)
					{
						ASSERT_DEBUG(wp <= sizeof(output));
						output[wp++] = MIN(sym_lit, 255);
						sym_lit -= 255;
					}

					while (lit_from < rp)
					{
						ASSERT_DEBUG(wp <= sizeof(output));
						output[wp++] = input[lit_from];
						lit_from++;
					}
				}

				int sym_match = match_len - RCM_MATCH_MIN;
				int sym_dist = match_dist - 1;
				if (match_is_short)
				{
					ASSERT_DEBUG(match_len - RCM_MATCH_MIN <= 8);
					ASSERT_DEBUG(sym_match < 8);

					byte token = 1 + (sym_match << 2) + ((sym_dist & 7) << 5);
					ASSERT_DEBUG(wp <= sizeof(output));
					output[wp++] = token;

					sym_dist >>= 3;
					ASSERT_DEBUG(sym_dist < 256);
					output[wp++] = sym_dist & 0xFF;
				}
				else
				{
					ASSERT_DEBUG(match_len - RCM_MATCH_MIN <= (1 << (6 + 8)));

					byte token = 2 + (match_len >= RCM_MATCH_MED_MAX) + (sym_match << 2);
					ASSERT_DEBUG(wp <= sizeof(output));
					output[wp++] = token;

					ASSERT_DEBUG(wp <= sizeof(output));
					output[wp++] = sym_dist;
					ASSERT_DEBUG(wp <= sizeof(output));
					output[wp++] = sym_dist >> 8;

					if (match_len >= RCM_MATCH_MED_MAX)
					{
						ASSERT_DEBUG(wp <= sizeof(output));
						output[wp++] = sym_match >> 6;
					}
				}

				lit_from = rp + match_len;
				while (rp < lit_from)
				{
					int h = hash4<HASH_BITS>(input + rp);

					++hash_head[h];
					hash_entry[(h << TRY_BITS) + (hash_head[h] & TRY_MASK)] = rp;

					rp++;
				}
			}
			else
			{
				int h = hash4<HASH_BITS>(input + rp);

				++hash_head[h];
				hash_entry[(h << TRY_BITS) + (hash_head[h] & TRY_MASK)] = rp;

				rp++;
			}
		}

		int sym_lit = rp - lit_from;
		if (sym_lit > 0)
		{
			ASSERT_DEBUG(wp <= sizeof(output));
			output[wp++] = MIN(sym_lit, 63) << 2;
			sym_lit -= 63;

			while (sym_lit >= 0)
			{
				ASSERT_DEBUG(wp <= sizeof(output));
				output[wp++] = MIN(sym_lit, 255);
				sym_lit -= 255;
			}

			while (lit_from < rp)
			{
				ASSERT_DEBUG(wp <= sizeof(output));
				output[wp++] = input[lit_from];
				lit_from++;
			}
		}

		fwrite(&wp, sizeof(int), 1, fout);
		fwrite(output, 1, wp, fout);

		memcpy(sliding_input, input, WINDOW_SIZE);
	}

	fclose(fin);
	fclose(fout);
}

void decompress(const char* infilename, const char* outfilename)
{
	FILE* fin = fopen(infilename, "rb");
	if (!fin)
	{
		printf("Could not open %s\n", infilename);
		return;
	}

	FILE* fout = fopen(outfilename, "wb");
	if (!fout)
	{
		fclose(fin);

		printf("Could not open %s\n", outfilename);
		return;
	}

	fseek(fin, 0, SEEK_END);
	uint64 inlen = ftell(fin);
	fseek(fin, 0, SEEK_SET);

	byte input[WINDOW_SIZE_WORST];
	byte output_mem[2 * WINDOW_SIZE + FAST_COPY_PADDING];

	memset(output_mem, 0, sizeof(output_mem));
	byte* output = output_mem + WINDOW_SIZE;
	byte* sliding_output = output_mem;

	bool failed = false;
	int n = 0;
	while (true)
	{
		int n = 0;
		fread(&n, sizeof(int), 1, fin);
		if (n == 0)
		{
			break;
		}

		if (n < 0 || n > WINDOW_SIZE_WORST)
		{
			printf("Error: n exceeds worst-case window size %d: %d\n", WINDOW_SIZE_WORST, n);
			failed = true;
			break;
		}

		if (fread(input, 1, n, fin) != n)
		{
			printf("Read incomplete block\n");
			failed = true;
			break;
		}

		int rp = 0, wp = 0;
		while (rp < n)
		{
			if (rp == n)
			{
				failed = true;
				break;
			}

			byte token = input[rp++];
			byte op = token & 3;
			if (op == 0)
			{
				int sym_lit = token >> 2;

				int num_literals = sym_lit;
				if (sym_lit == 63)
				{
					while (sym_lit >= 0)
					{
						if (rp == n)
						{
							failed = true;
							break;
						}

						sym_lit = input[rp++];
						num_literals += sym_lit;
						sym_lit -= 255;
					}
				}

				if (rp + num_literals > n || wp + num_literals > WINDOW_SIZE)
				{
					failed = true;
					break;
				}

				string_copy<false>(input + rp, output + wp, num_literals);

				rp += num_literals;
				wp += num_literals;

				ASSERT_DEBUG(wp <= WINDOW_SIZE);
			}
			else
			{
				int len, dist;
				if (op == 1)
				{
					len = (token >> 2) & 7;
					dist = (token >> 5) & 7;

					if (rp == n)
					{
						failed = true;
						break;
					}

					int sym_dist = input[rp++];
					dist += sym_dist << 3;
				}
				else if (op & 0x2)
				{
					len = (token >> 2) & 63;

					if (rp == n)
					{
						failed = true;
						break;
					}

					dist = input[rp++];

					if (rp == n)
					{
						failed = true;
						break;
					}

					dist += input[rp++] << 8;

					if (op == 3)
					{
						if (rp == n)
						{
							failed = true;
							break;
						}

						len += input[rp++] << 6;
					}
				}
				else
				{
					failed = true;
					break;
				}

				len += RCM_MATCH_MIN;
				dist++;

				if (dist > WINDOW_SIZE)
				{
					failed = true;
					break;
				}

				if (len > 0)
				{
					if (wp + len > WINDOW_SIZE)
					{
						failed = true;
						break;
					}

					int mp = wp - dist + WINDOW_SIZE;
					ASSERT_DEBUG(mp >= 0 && mp < 2 * WINDOW_SIZE);

					string_copy<true>(sliding_output + mp, output + wp, len);
					wp += len;

					ASSERT_DEBUG(wp <= WINDOW_SIZE);
				}
			}
		}

		if (failed)
		{
			break;
		}

		ASSERT_DEBUG(wp <= WINDOW_SIZE);
		fwrite(output, 1, wp, fout);
		memcpy(sliding_output, output, WINDOW_SIZE);
	}

	if (failed)
	{
		printf("Decompression failed\n");
	}

	fclose(fin);
	fclose(fout);
}

void compress_optimal(const char* infilename, const char* outfilename)
{
	const int DICT_ROTATE_MASK = (1 << 30) - 1;

	const static int HASH_BITS = 15;
	const static int HASH_SIZE = 1 << HASH_BITS;

	FILE* fin = fopen(infilename, "rb");
	if (!fin)
	{
		printf("Could not open %s\n", infilename);
		return;
	}

	FILE* fout = fopen(outfilename, "wb");
	if (!fout)
	{
		fclose(fin);

		printf("Could not open %s\n", outfilename);
		return;
	}

	fseek(fin, 0, SEEK_END);
	uint64 inlen = ftell(fin);
	fseek(fin, 0, SEEK_SET);

	uint32 hash_head[HASH_SIZE], hash_next[WINDOW_SIZE][2];

	struct
	{
		uint32 cost;
		uint16 dists[2];
		uint16 max_length_short, max_length;
	} path[WINDOW_SIZE + 1];

	byte input[2 * WINDOW_SIZE], output[WINDOW_SIZE_WORST];

	memset(hash_head, -1, sizeof(hash_head));
	memset(hash_next, -1, sizeof(hash_next));

	int n = 0;
	int offset_n = 0;
	while ((n = fread(input, 1, WINDOW_SIZE, fin)) > 0)
	{
		int rp = 0;
		while (rp < n)
		{
			path[rp].max_length_short = RCM_MATCH_MIN - 1;
			path[rp].max_length = RCM_MATCH_MIN - 1;

			int max_match_len = MIN(RCM_MATCH_LONG_MAX - 1, n - rp);
			max_match_len = MAX(0, max_match_len - RCM_MATCH_SEARCH_PADDING);
			if (max_match_len >= RCM_MATCH_MIN)
			{
				uint32* left = &hash_next[rp & WINDOW_MASK][1];
				uint32* right = &hash_next[rp & WINDOW_MASK][0];

				int left_len = 0, right_len = 0;

				int hash = hash4<HASH_BITS>(input + rp);
				uint32 sp = hash_head[hash];
				hash_head[hash] = rp + offset_n;

				while (sp != uint32(-1) && rp + offset_n - sp < WINDOW_SIZE)
				{
					uint32 mp = sp & WINDOW_MASK;
					if (mp > rp)
					{
						mp += WINDOW_SIZE;
					}

					int len = MIN(left_len, right_len);
					if (input[(mp + len) & WINDOW_DOUBLE_MASK] == input[rp + len])
					{
						while (len < max_match_len && input[(mp + len) & WINDOW_DOUBLE_MASK] == input[rp + len])
						{
							++len;
						}

						int match_dist = (rp - mp) & WINDOW_MASK;
						if (len > path[rp].max_length_short && match_dist <= RCM_OFFSET_THRESHOLD)
						{
							path[rp].dists[0] = match_dist;
							path[rp].max_length_short = MIN(len, RCM_MATCH_THRESHOLD - 1);
						}

						if (len > path[rp].max_length)
						{
							path[rp].dists[1] = match_dist;
							path[rp].max_length = len;

							if (len == max_match_len)
							{
								break;
							}
						}
					}

					if (input[(mp + len) & WINDOW_DOUBLE_MASK] < input[rp + len])
					{
						*right = sp;
						right = &hash_next[sp & WINDOW_MASK][1];
						sp = *right;
						right_len = len;
					}
					else
					{
						*left = sp;
						left = &hash_next[sp & WINDOW_MASK][0];
						sp = *left;
						left_len = len;
					}
				}

				*left = -1;
				*right = -1;
			}

			if (path[rp].max_length > RCM_MATCH_SEMIOPTIMAL_MAX + RCM_MATCH_MIN)
			{
				int sp = rp;
				for (int skip = path[rp].max_length - RCM_MATCH_SEMIOPTIMAL_MAX - RCM_MATCH_MIN;
					skip > 0 && rp < n; skip--, rp++)
				{
					path[rp].max_length_short = MAX(0, path[sp].max_length_short - (rp - sp));
					path[rp].max_length = MAX(0, path[sp].max_length - (rp - sp));
					//printf("%d\n", path[sp].max_length - (rp - sp));
					path[rp].dists[0] = path[sp].dists[0];
					path[rp].dists[1] = path[sp].dists[1];
				}
			}
			else
			{
				++rp;
			}
		}

		rp = 0;
		int wp = 0;
		int lit_from = 0;

		path[n].cost = 0;
		int literal_count = 0;
		for (int p = n - 1; p >= 0; p--)
		{
			path[p].cost = uint32(-1);

			int cost_next = path[p + 1].cost + 1;
			if (literal_count == 0)
			{
				++cost_next;
			}
			else if (literal_count == 63 || (literal_count > 63 && ((literal_count - 63) % 255) == 0))
			{
				++cost_next;
			}

			literal_count++;
			if (path[p].max_length < RCM_MATCH_MIN)
			{
				path[p].cost = cost_next;
			}
			else
			{
				for (int i = path[p].max_length; i >= RCM_MATCH_MIN; i--)
				{
					int cost_match = path[p + i].cost;
					if (i <= path[p].max_length_short)
					{
						cost_match += 2;
					}
					else if (i < RCM_MATCH_MED_MAX)
					{
						cost_match += 3;
					}
					else
					{
						cost_match += 4;
					}

					if (cost_match < path[p].cost)
					{
						path[p].cost = cost_match;
						path[p].max_length = i;
					}
				}

				if (cost_next < path[p].cost)
				{
					path[p].cost = cost_next;
					path[p].max_length = 0;
				}
				else
				{
					literal_count = 0;
				}
			}
		}

		while (rp < n)
		{
			int match_len = path[rp].max_length;
			if (match_len >= RCM_MATCH_MIN)
			{
				int match_is_short = match_len <= path[rp].max_length_short;
				int match_dist = path[rp].dists[1 - match_is_short];

				int sym_lit = rp - lit_from;
				if (sym_lit > 0)
				{
					ASSERT_DEBUG(wp <= sizeof(output));

					byte token = MIN(sym_lit, 63) << 2;
					output[wp++] = token;
					sym_lit -= 63;

					while (sym_lit >= 0)
					{
						ASSERT_DEBUG(wp <= sizeof(output));
						output[wp++] = MIN(sym_lit, 255);
						sym_lit -= 255;
					}

					while (lit_from < rp)
					{
						ASSERT_DEBUG(wp <= sizeof(output));
						output[wp++] = input[lit_from];
						lit_from++;
					}
				}

				int sym_match = match_len - RCM_MATCH_MIN;
				int sym_dist = match_dist - 1;
				if (match_is_short)
				{
					ASSERT(match_len - RCM_MATCH_MIN <= 8);
					ASSERT(sym_match < 8);

					byte token = 1 + (sym_match << 2) + ((sym_dist & 7) << 5);
					ASSERT_DEBUG(wp <= sizeof(output));
					output[wp++] = token;

					sym_dist >>= 3;
					ASSERT(sym_dist < 256);
					output[wp++] = sym_dist;
				}
				else
				{
					ASSERT_DEBUG(match_len - RCM_MATCH_MIN <= (1 << (6 + 8)));

					byte token = 2 + (match_len >= RCM_MATCH_MED_MAX) + (sym_match << 2);
					ASSERT_DEBUG(wp <= sizeof(output));
					output[wp++] = token;

					ASSERT_DEBUG(wp <= sizeof(output));
					output[wp++] = sym_dist;
					ASSERT_DEBUG(wp <= sizeof(output));
					output[wp++] = sym_dist >> 8;

					if (match_len >= RCM_MATCH_MED_MAX)
					{
						output[wp++] = sym_match >> 6;
					}
				}

				lit_from = rp + match_len;
				rp = lit_from;
			}
			else
			{
				rp++;
			}

		}

		int sym_lit = rp - lit_from;
		if (sym_lit > 0)
		{
			ASSERT_DEBUG(wp <= sizeof(output));
			output[wp++] = MIN(sym_lit, 63) << 2;
			sym_lit -= 63;

			while (sym_lit >= 0)
			{
				ASSERT_DEBUG(wp <= sizeof(output));
				output[wp++] = MIN(sym_lit, 255);
				sym_lit -= 255;
			}

			while (lit_from < rp)
			{
				ASSERT_DEBUG(wp <= sizeof(output));
				output[wp++] = input[lit_from];
				lit_from++;
			}
		}

		fwrite(&wp, sizeof(int), 1, fout);
		fwrite(output, 1, wp, fout);
		memcpy(input + WINDOW_SIZE, input, WINDOW_SIZE);

		offset_n += n;
		if (offset_n & ~DICT_ROTATE_MASK)
		{
			offset_n = 0;

			memset(hash_head, -1, sizeof(hash_head));
			memset(hash_next, -1, sizeof(hash_next));
		}
	}

	//printf("%d (%d KB) -> %d (%d KB)\n", ftell(fin), (ftell(fin) + 1023) >> 10, ftell(fout), (ftell(fout) + 1023) >> 10);
	fclose(fin);
	fclose(fout);
}

struct Timer
{
	void Start();
	void Stop();
	double Elapsed();
	double ElapsedSec();

	clock_t tstart, tstop;
};

void Timer::Start()
{
	tstart = clock();
}

void Timer::Stop()
{
	tstop = clock();
}

double Timer::Elapsed()
{
	const double div = 1000.0 / CLOCKS_PER_SEC;
	return (tstop - tstart) * div;
}

double Timer::ElapsedSec()
{
	const double div = 1.0 / CLOCKS_PER_SEC;
	return (tstop - tstart) * div;
}

static int tolower(int c)
{
	if (c >= 'A' && c <= 'Z')
	{
		return 'a' + (c - 'A');
	}

	return c;
}

int main(int argc, const char** argv)
{
	FILE* f = 0;
	if (argc == 4 && tolower(argv[1][0]) == 'c')
	{
		if ((f = fopen(argv[3], "rb")))
		{
			printf("Error: %s already exists\n", argv[3]);
			return 0;
		}

		clock_t tspan = clock();
		if (tolower(argv[1][1] == 'o'))
		{
			compress_optimal(argv[2], argv[3]);
		}
		else
		{
			compress_fast(argv[2], argv[3]);
		}
		tspan = clock() - tspan;
		printf("Completed in %.3f sec\n", tspan / float(CLOCKS_PER_SEC));
	}
	else if (argc == 4 && tolower(argv[1][0]) == 'd')
	{
		if ((f = fopen(argv[3], "rb")))
		{
			printf("Error: %s already exists\n", argv[3]);
			return 0;
		}

		clock_t tspan = clock();
		decompress(argv[2], argv[3]);
		tspan = clock() - tspan;
		printf("Completed in %.3f sec\n", tspan / float(CLOCKS_PER_SEC));
	}
	else
	{
		printf("RCM - Written by Nauful\n"
			"Options:\n"
			"\tc [input] [output] - Compress input file to output file (fast)\n"
			"\tco [input] [output] - Compress input file to output file (best)\n"
			"\td [input] [output] - Decompress input file to output file\n"
		);
	}

	return 0;
}
